import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

version(Win64){
  ushort HIWORD(ulong l){ return cast(WORD)(cast(DWORD_PTR)(l) >>> 16);}
  ushort LOWORD(ulong l){ return cast(WORD)(cast(DWORD_PTR)(l) & 0xffff); }
}else{
  ushort LOWORD(uint l){ return l & 0xffff; }
  ushort HIWORD(uint l){ return l >>> 16; }
}
COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0001) ;
const RBUTTONUP     = (0x0002) ;
const RBUTTONDBLCLK = (0x0004) ;
const ALT_KEY       = (0x0001) ;
const CNTL_KEY      = (0x0002) ;
const SHIFT_KEY     = (0x0004) ;

class GraphicBase : WindowObject {
  HDC        WinDC;
  HDC        WinDC_back;
  HBITMAP    bmp_back  ;
  BITMAPINFO bi ;
  SCROLLINFO rollinfo_H ;
  SCROLLINFO rollinfo_V ;
  int        rollstep   ;
  int IBwidth =0 ;
  int IBheight=0 ;
  int mouseX  =0 ;
  int mouseY  =0 ;
  int mouseB  =0 ;
  int ACS_stat=0 ;
  uint  *vram ;
  

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
    vram = null ;
    rollstep=20 ;
  }

  int setImageSize(int width, int height){
    if( width>0 && height>0 ){
      IBwidth =width  ;
      IBheight=height ;
      if( getHWND() != null ){
	DeleteDC(WinDC_back);
        DeleteObject(bmp_back);
	bi.bmiHeader.biWidth            =  IBwidth  ;
	bi.bmiHeader.biHeight           = -IBheight ;
	WinDC_back=CreateCompatibleDC(WinDC);
	bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&vram), null, 0 );
	SelectObject(WinDC_back, bmp_back);
	reDisplay() ;
      }
    }else{
      IBwidth =0      ;
      IBheight=0      ;
      if( getHWND() != null ){
	DeleteDC(WinDC_back);
	DeleteObject(bmp_back);
	reDisplay() ;
      }
    }
    return(0) ;
  }

  int reDisplay(){
    if( getHWND() != null ){
      SendMessage( getHWND(), WM_SIZE, 0, (winH<<16 | winW) );
      RedrawWindow( getHWND(), null, null, RDW_INVALIDATE ) ;
    }
    return(0) ;
  }
  
  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      this.wm_destroy() ;
      DeleteDC(WinDC);
      DeleteDC(WinDC_back);
      DeleteObject(bmp_back);
      break ;
      
    case WM_CREATE:
      bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
      bi.bmiHeader.biWidth            =  IBwidth  ;
      bi.bmiHeader.biHeight           = -IBheight ;
      bi.bmiHeader.biPlanes           = 1;
      bi.bmiHeader.biBitCount         = 32;
      bi.bmiHeader.biCompression      = BI_RGB;
      bi.bmiHeader.biSizeImage        = 0;
      bi.bmiHeader.biXPelsPerMeter    = 0;
      bi.bmiHeader.biYPelsPerMeter    = 0;
      bi.bmiHeader.biClrUsed          = 1;
      bi.bmiHeader.biClrImportant     = 0;

      rollinfo_H.cbSize = SCROLLINFO.sizeof ;
      rollinfo_H.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE ;
      rollinfo_H.nPos   = 0         ;
      rollinfo_H.nPage  = winW-1    ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = IBwidth-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      
      rollinfo_V.cbSize = SCROLLINFO.sizeof ;
      rollinfo_V.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE ;
      rollinfo_V.nPos   = 0          ;
      rollinfo_V.nPage  = winH-1     ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = IBheight-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;

      RECT wrect ;
      GetClientRect( hwnd, &wrect ) ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_H.nPage  = wrect.right;
      rollinfo_H.nMin   = 0          ;
      rollinfo_H.nMax   = IBwidth-1  ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_V.nPage  = wrect.bottom;
      rollinfo_V.nMin   = 0           ;
      rollinfo_V.nMax   = IBheight-1  ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      
      WinDC=GetDC(hwnd);
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&vram), null, 0 );
      SelectObject(WinDC_back, bmp_back);

      mouseB=0 ;
      this.wm_create() ;
      break ;

    case WM_HSCROLL:
      {
	int changed=1 ;
	
	switch( LOWORD(wp) ){
	case SB_LINELEFT :
	  rollinfo_H.nPos -=rollstep ;
	  if( rollinfo_H.nPos < 0 ){
	    rollinfo_H.nPos=0 ;
	  }
	  break ;
	case SB_LINERIGHT :
	  rollinfo_H.nPos +=rollstep ;
	  if( rollinfo_H.nPos > rollinfo_H.nMax ){
	    rollinfo_H.nPos=rollinfo_H.nMax ;
	  }
	  break ;
	case SB_PAGELEFT :
	  rollinfo_H.nPos -= rollinfo_H.nPage/2 ;
	  if( rollinfo_H.nPos < 0 ){
	    rollinfo_H.nPos=0 ;
	  }
	  break ;
	case SB_PAGERIGHT :
	  rollinfo_H.nPos += rollinfo_H.nPage/2 ;
	  if( rollinfo_H.nPos > rollinfo_H.nMax ){
	    rollinfo_H.nPos=rollinfo_H.nMax ;
	  }
	  break ;
	case SB_THUMBTRACK :
	  {
	    SCROLLINFO  tmp_scrinfo ;
	    tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
	    GetScrollInfo( hwnd, SB_HORZ, &tmp_scrinfo) ;
	    rollinfo_H.nPos = tmp_scrinfo.nTrackPos ;
	  }
	  break ;
	  
	default:
	  changed=0 ;
	  break ;
	}
      
	if( changed!=0 ){
	  SetScrollInfo( hwnd, SB_HORZ, &rollinfo_H, true) ;
	  InvalidateRect(hwnd , null , true);
	}
      }
      break ;
      
    case WM_VSCROLL:
      {
	int changed=1 ;
	
	switch( LOWORD(wp) ){
	case SB_LINEUP :
	  rollinfo_V.nPos -=rollstep ;
	  if( rollinfo_V.nPos < 0 ){
	    rollinfo_V.nPos=0 ;
	  }
	  break ;
	case SB_LINEDOWN :
	  rollinfo_V.nPos +=rollstep ;
	  if( rollinfo_V.nPos > rollinfo_V.nMax ){
	    rollinfo_V.nPos=rollinfo_V.nMax ;
	  }
	  break ;
	case SB_PAGEUP :
	  rollinfo_V.nPos -= rollinfo_V.nPage/2 ;
	  if( rollinfo_V.nPos < 0 ){
	    rollinfo_V.nPos=0 ;
	  }
	  break ;
	case SB_PAGEDOWN :
	  rollinfo_V.nPos += rollinfo_V.nPage/2 ;
	  if( rollinfo_V.nPos > rollinfo_V.nMax ){
	    rollinfo_V.nPos=rollinfo_V.nMax ;
	  }
	  break ;
	case SB_THUMBTRACK :
	  {
	    SCROLLINFO  tmp_scrinfo ;
	    tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
	    GetScrollInfo( hwnd, SB_VERT, &tmp_scrinfo) ;
	    rollinfo_V.nPos = tmp_scrinfo.nTrackPos ;
	  }
	  break ;
	  
	default:
	  changed=0 ;
	  break ;
	}
	if( changed!=0 ){
	  SetScrollInfo( hwnd, SB_VERT, &rollinfo_V, true) ;
	  InvalidateRect(hwnd , null , true);
	}
      }
      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC         hdc;
      SCROLLINFO  slinfo_H ;
      SCROLLINFO  slinfo_V ;
      HBRUSH      COLBG  ;
      HPEN        PCOLBG ;

      slinfo_H.cbSize = SCROLLINFO.sizeof ; slinfo_H.fMask = SIF_POS ;
      slinfo_V.cbSize = SCROLLINFO.sizeof ; slinfo_V.fMask = SIF_POS ;
      GetScrollInfo( hwnd, SB_HORZ, &slinfo_H) ;
      GetScrollInfo( hwnd, SB_VERT, &slinfo_V) ;
      
      PCOLBG = CreatePen(PS_SOLID, 1, RGBA(128,128,128,0) & 0x00ffffff) ;
      COLBG  = CreateSolidBrush(RGBA(128,128,128,0) & 0x00ffffff) ;
      hdc = BeginPaint(hwnd, &ps);
      SelectObject(hdc, PCOLBG) ;
      SelectObject(hdc, COLBG) ;
      if( IBwidth<=0 || IBheight<=0 ){
	Rectangle(hdc, 0,0,winW,winH) ;
      }else{
	StretchBlt(hdc, 0-slinfo_H.nPos, 0-slinfo_V.nPos, IBwidth, IBheight, WinDC_back, 0, 0, IBwidth, IBheight, SRCCOPY);
	if( (winW - IBwidth)>0 ){
	  Rectangle(hdc, IBwidth,0,winW,winH) ;
	}
	if( (winH - IBheight)>0 ){
	  Rectangle(hdc, 0,IBheight,winW,winH) ;
	}
      }
      EndPaint(hwnd, &ps);
      DeleteObject(PCOLBG) ;
      DeleteObject(COLBG) ;
      break ;

    case WM_MOVE:
      return(0) ;
      break ;

    case WM_WINDOWPOSCHANGING :
      rollinfo_H.fMask  = SIF_RANGE ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = IBwidth-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = IBheight-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      break ;
      
    case WM_SIZE:
      int w=LOWORD(lp) ;
      int h=HIWORD(lp) ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_H.nPage  = w         ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = IBwidth-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_V.nPage  = h          ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = IBheight-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;

      RECT wrect ;
      GetClientRect( hwnd, &wrect ) ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_H.nPage  = wrect.right;
      rollinfo_H.nMin   = 0          ;
      rollinfo_H.nMax   = IBwidth-1  ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE ;
      rollinfo_V.nPage  = wrect.bottom;
      rollinfo_V.nMin   = 0           ;
      rollinfo_V.nMax   = IBheight-1  ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      
      this.wm_size(w,h) ;
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      {
	POINT point;
	SCROLLINFO slinfo_H ;
	SCROLLINFO slinfo_V ;

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );

	slinfo_H.cbSize = SCROLLINFO.sizeof ; slinfo_H.fMask = SIF_TRACKPOS ;
	slinfo_V.cbSize = SCROLLINFO.sizeof ; slinfo_V.fMask = SIF_TRACKPOS ;
	GetScrollInfo( hwnd, SB_HORZ, &slinfo_H) ;
	GetScrollInfo( hwnd, SB_VERT, &slinfo_V) ;
	
	mouseX = point.x + slinfo_H.nTrackPos ;
	mouseY = point.y + slinfo_V.nTrackPos ;
	if( msg==WM_LBUTTONDOWN || msg==WM_RBUTTONDOWN ) SetCapture( hwnd ) ;
	if( msg==WM_LBUTTONUP   || msg==WM_RBUTTONUP   ) ReleaseCapture() ;
	if( msg==WM_MOUSEMOVE ){
	}else{
	  mouseB=0 ;
	  switch( msg ){
	  case WM_LBUTTONDOWN:
	    mouseB|=LBUTTONDOWN ;
	    break ;
	  case WM_LBUTTONUP:
	    mouseB|=LBUTTONUP ;
	    break ;
	  case WM_LBUTTONDBLCLK:
	    mouseB|=LBUTTONDBLCLK ;
	    break ;
	  case WM_RBUTTONDOWN:
	    mouseB|=RBUTTONDOWN ;
	    break ;
	  case WM_RBUTTONUP:
	    mouseB|=RBUTTONUP ;
	    break ;
	  case WM_RBUTTONDBLCLK:
	    mouseB|=RBUTTONDBLCLK ;
	    break ;
	  default : break ;
	  }	  
	  
	  if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	  else                            ACS_stat &= ~(SHIFT_KEY) ;
	  if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	  else                            ACS_stat &= ~(CNTL_KEY)  ;
	  if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	  else                            ACS_stat &= ~(ALT_KEY)   ;
	  
	}
	if( this.mouse(ACS_stat, mouseB, mouseX, mouseY )!=0 ){
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int wm_destroy(){
    return(0) ;
  }
  int wm_create(){
    return(0) ;
  }
  int wm_paint(){
    return(0) ;
  }
  int wm_size(int w, int h){
    return(0) ;
  }
  
  uint* getImageBuffer(){
    return(vram) ;
  }

  int getImageBufferWidth(){
    return(IBwidth) ;
  }
  int getImageBufferHeight(){
    return(IBheight) ;
  }

  int mouse( int acs_stat, int mouseb, int mousex, int mousey ){
    return(0) ;
  }

}
