import win32.windows;
import win32.windef;
import win32.commctrl;

import std.stdio;
import std.string;

import util.Image;

class BitMap
{

  uint[]   imgbuf ;
  int      w,h ;
  
  this(){
  }

  ~this(){
  }

  int getWidth(){
    return(w) ;
  }

  int getHeight(){
    return(h) ;
  }
  
  int readImage(HINSTANCE hinstance, const wchar* name, const wchar* type){
    HRSRC   hrs = FindResource(hinstance, name, type) ;
    if( hrs!=null ){
      int     size    = SizeofResource(hinstance, hrs);
      HGLOBAL hglobal = LoadResource(hinstance, hrs);
      if( hglobal!=null ){
	ubyte *pdata = cast(ubyte*)(LockResource(hglobal)) ;
	ubyte mdata[] ; mdata = pdata[0..size] ;
	auto image = new Image() ;
	image.open(mdata, "png") ;
	w = image.getWidth()  ;
	h = image.getHeight() ;
	imgbuf.length = w*h ;
	for( int i=0 ; i<h ; i++ ){
	  uint[] img = image.readline(1,true) ;
	  imgbuf[w*i .. w*i+w]=img[0..$] ;
	  delete img ;
	}
	image.close() ;
	FreeResource(hglobal);
	return(0) ;
      }
    }
    return(-1) ;
  }

  HBITMAP makeHBITMAP(uint bgcol){
    HBITMAP    hbitmap ;
    uint       *bmpp ;
    BITMAPINFO bi ;
    
    bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
    bi.bmiHeader.biWidth            =  w ;
    bi.bmiHeader.biHeight           = -h ;
    bi.bmiHeader.biPlanes           = 1;
    bi.bmiHeader.biBitCount         = 32;
    bi.bmiHeader.biCompression      = BI_RGB;
    bi.bmiHeader.biSizeImage        = 0;
    bi.bmiHeader.biXPelsPerMeter    = 0;
    bi.bmiHeader.biYPelsPerMeter    = 0;
    bi.bmiHeader.biClrUsed          = 1;
    bi.bmiHeader.biClrImportant     = 0;
    hbitmap = CreateDIBSection( null, &bi, DIB_RGB_COLORS, cast(void**)(&bmpp), null, 0 );
    
    for( int i=0 ; i<h ; i++ ){
      bmpp[w*i .. w*i+w]=imgbuf[w*i .. w*i+w] ;
      for( int j=0 ; j<w ; j++ ){
	int br = ((bgcol & 0x00ff0000)>>16) ;
	int bg = ((bgcol & 0x0000ff00)>>8 ) ;
	int bb = ((bgcol & 0x000000ff)    ) ;
	int fa = (bmpp[w*i+j] & 0xff000000)>>24 ;
	int fr = (((bmpp[w*i+j] & 0x000000ff)    ) * fa + br*(255-fa))/255 ;
	int fg = (((bmpp[w*i+j] & 0x0000ff00)>>8 ) * fa + bg*(255-fa))/255 ;
	int fb = (((bmpp[w*i+j] & 0x00ff0000)>>16) * fa + bb*(255-fa))/255 ;
	bmpp[w*i+j] = 0xff000000 | (fr &0xff) | (fg &0xff)<<8 | (fb &0xff)<<16 ;
      }
    }
    return(hbitmap) ;
  }
}  
